import os

import pytest

from wcs import fields
from wcs.admin.settings import UserFieldsFormDef
from wcs.carddef import CardDef
from wcs.qommon.http_request import HTTPRequest

from ..utilities import clean_temporary_pub, create_temporary_pub, get_app
from .utils import sign_uri


@pytest.fixture
def pub(emails):
    pub = create_temporary_pub()

    req = HTTPRequest(None, {'SCRIPT_NAME': '/', 'SERVER_NAME': 'example.net'})
    pub.set_app_dir(req)
    pub.cfg['identification'] = {'methods': ['password']}
    pub.cfg['language'] = {'language': 'en'}
    pub.write_cfg()

    with open(os.path.join(pub.app_dir, 'site-options.cfg'), 'w') as fd:
        fd.write(
            '''\
[api-secrets]
coucou = 1234
'''
        )

    pub.user_class.wipe()

    return pub


def teardown_module(module):
    clean_temporary_pub()


def test_carddata_user_fields(pub, local_user):
    pub.role_class.wipe()
    role = pub.role_class(name='test')
    role.store()

    local_user.roles = [role.id]
    local_user.store()

    CardDef.wipe()
    carddef = CardDef()
    carddef.name = 'test'
    carddef.workflow_roles = {'_viewer': role.id}
    carddef.store()

    carddata = carddef.data_class()()
    carddata.just_created()
    carddata.store()

    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert 'user' not in resp.json
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert 'user' not in resp.json['data'][0]

    carddata.user_id = local_user.id
    carddata.store()
    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert resp.json['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
    }
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert resp.json['data'][0]['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
    }

    formdef = UserFieldsFormDef(pub)
    formdef.fields = [
        fields.StringField(id='3', label='test', type='string', varname='var1'),
        fields.StringField(id='9', label='noop', type='string', varname='var2'),
        fields.StringField(id='42', label='no varname', type='string'),
    ]
    formdef.store()

    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert resp.json['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': None,
        'var2': None,
    }
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert resp.json['data'][0]['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': None,
        'var2': None,
    }

    local_user.form_data = {'3': 'toto', '9': 'nono'}
    local_user.store()
    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert resp.json['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': 'toto',
        'var2': 'nono',
    }
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert resp.json['data'][0]['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': 'toto',
        'var2': 'nono',
    }
